#include <stdio.h>
#include <stdlib.h>
#include <time.h>

//---------------------------------------------------------

void test01( void ) {         // расширение GCC
   int array[] = { 1, -2, 3, -4, 5, -6, 7 },
       size = sizeof( array ) / sizeof( *array ), i;
   void pow2( void ) {        // вложенное описание функции pow2()
      int decr( int arg ) {   // ещё один уровень вложенности функции decr()
         return arg - 1;
      }
      for( i = 0; i < size; i++ )
         array[ i ] = decr( array[ i ] ) * decr( array[ i ] );
   }
   printf( "вложенные функции GCC:\n" );
   printf( "до\t:" );
   for( i = 0; i < size; i++ )
      printf( "%2d%s", array[ i ], ( i == size - 1 ? "\n" : " , " ) );
   pow2();
   printf( "после\t:" );
   for( i = 0; i < size; i++ )
      printf( "%2d%s", array[ i ], ( i == size - 1 ? "\n" : " , " ) );
}

//---------------------------------------------------------

static void show( int *arr, int row, int col ) {
   int r, c;
   printf( "матрица : %d x %d [ %d ]\n",
           row, col, row * col );
   for( r = 0; r < row; r++ ) {
      for( c = 0; c < col; c++ )
         printf( "%3d", arr[ r * col + c ] );
      printf( "\n" );
   }
}

static void transpa( int *arr, int *row, int *col ) {
   int r, c;
   int wrk[ *row * *col ];    // массив с динамической размерностью
   for( r = 0; r < *row; r++ )
      for( c = 0; c < *col; c++ ) {
         int i1 = r * *col + c,
             i2 = c * *row + r;
         wrk[ i2 ] = arr[ i1 ];
      }
   for( r = 0; r < *row * *col; r++ ) {
      arr[ r ] = wrk[ r ];
   }
   r = *row;
   *row = *col;
   *col = r;
}

#define COL 5
#define ROW 2

void test02( void ) {           // динамические массивы (C99)
   int c[ ROW ][ COL ] = {
          { 1, 2, 3, 4, 5 },
          { 2, 3, 4, 5, 6 }
       },
       col = COL, row = ROW;
   show( (int*)c, row, col );
   printf( "транспонирование не квадратной матрицы:\n" ); 
   transpa( (int*)c, &row, &col );
   show( (int*)c, row, col );
}

//---------------------------------------------------------

static unsigned long long rdtsc( void ) {
   unsigned long long int x;
   asm volatile ( "rdtsc" : "=A" (x) ); // команда RDTSC
   return x;
}

void test03( void ) {                   // ассемблерные вставки (GCC)
   time_t t1, t2;
   unsigned long long cf, cs; 
   time( &t1 );
   while( t1 == time( &t2 ) ) cf  = rdtsc(); // начало очередной секунды
   while( t2 == time( &t1 ) ) cs  = rdtsc(); // завершение этой секунды
   printf( "тактовая частота процессора %.3f Ghz\n",
           (double)( cs - cf ) / 1.E9 ); 
}

//---------------------------------------------------------

typedef struct vararr {
   int n, data[ 0 ];
} vararr_t;

void varfunc( vararr_t *a ) {
   int ni = a->n, j;
   printf( "массив размера %d\t=> { ", ni );
   int *va = (int*)a;
   for( j = 1; j <= ni; j++ )
      printf( "%d%s", va[ j ], j != ni ? " , " : " }\n" );
}

void test04( void ) {
   printf( "структуры переменного размера:\n" );
   int var[] = { 3, 5, 7, 10 }, i;
   for( i = 0; i < sizeof( var ) / sizeof( *var ); i++ ) {
      int len = var[ i ], j;
      vararr_t *arr = (vararr_t*)calloc( len + 1, sizeof( int ) );
      arr->n = len;
      for( j = 0; j < len; j++ ) arr->data[ j ] = j + 1;
      varfunc( arr );
      free( arr );
   }
}

//---------------------------------------------------------

void ( *tests[] )( void ) = {          // последовательность тестов
        test01,                        // вложенные описания функций
        test02,                        // массивы динамической размерности
        test03,                        // ассемблерные вставки
        test04,                        // структуры переменного размера
};
#include "main.c"

